import matplotlib.pyplot as plt

from .. import erfani_decomp_func as ed  # noqa: E402

PANELS1 = ["a)", "b)", "c)", "d)", "e)", "f)", "g)", "h)"]
marker10 = ["o", "s", "d", "^", "+", "x", "*", "p"]
color2 = ["k", "b", "tab:red", "tab:green", "k", "b", "tab:red", "tab:green"]


def fig13(res, lb):
    r_positive = res.loc[
        (slice(None), slice(None), slice(None), slice(None), slice(None), 1)
    ]
    r_negative = res.loc[
        (slice(None), slice(None), slice(None), slice(None), slice(None), -1)
    ]
    assert len(r_positive) == len(r_negative), "Datasets must have same length."

    fig, axis = plt.subplots(2, 2, figsize=(12, 10))

    metadata = {
        (0, 0): {"ylabel": "$ΔA_c$", "title": "L06"},
        (0, 1): {"ylabel": None, "title": "L10"},
        (1, 0): {"ylabel": "$-ΔR$ ($W$ $m^{-2}$)", "title": None},
        (1, 1): {"ylabel": None, "title": None},
    }

    for i, j in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        axis[i, j].set_ylabel(metadata[(i, j)]["ylabel"], fontsize="18")
        axis[i, j].set_title(metadata[(i, j)]["title"], fontsize="23")
        axis[i, j].set_xlim([0.5, 10])
        if i == 0:
            axis[i, j].set_ylim([-0.025, 0.1])
        else:
            axis[i, j].set_ylim([-35, 120])
        axis[i, j].tick_params(axis="both", which="major", labelsize=14)
        axis[i, j].grid(linestyle=":", linewidth=1)  # axis='y',
        axis[i, j].text(
            0.1,
            1.08,
            PANELS1[i * 2 + j],
            transform=axis[i, j].transAxes,
            fontsize=18,
            fontweight="bold",
            va="top",
            fontfamily="serif",
            ha="right",
        )

    #####
    for ii in range(len(r_positive)):
        rp = r_positive.iloc[ii]
        rn = r_negative.iloc[ii]

        (
            rN_nan_fd,
            rL_nan_fd,
            rC_nan_fd,
            CRE_M_fd,
            CRE_T_fd,
            CRE_L_fd,
            CRE_CF_fd,
            CRE_rs_fd,
            A_M_fd,
            A_T_fd,
            A_L_fd,
            A_CF_fd,
            A_rs_fd,
            CRE_M_fd1,
            CRE_T_fd1,
            CRE_L_fd1,
            CRE_CF_fd1,
            CRE_rs_fd1,
            A_M_fd1,
            A_T_fd1,
            A_L_fd1,
            A_CF_fd1,
            A_rs_fd1,
            CRE_M_fd2,
            CRE_T_fd2,
            CRE_L_fd2,
            CRE_CF_fd2,
            CRE_rs_fd2,
            A_M_fd2,
            A_T_fd2,
            A_L_fd2,
            A_CF_fd2,
            A_rs_fd2,
            CF1_nan_fd,
            CF2_nan_fd,
            CF_nan_fd,
        ) = ed.quant_CRE_avg(rp, rn, output="extended")

        ####
        i = 0
        if ii < len(r_positive) // 2:
            j = 0
        else:
            j = 1
        axis[i, j].errorbar(
            rN_nan_fd,
            A_T_fd,
            yerr=abs(A_T_fd2 - A_T_fd1) / 2,
            markersize=10,
            fmt=marker10[1],
            color=color2[ii],
            mfc="white",
        )
        axis[i, j].errorbar(
            rN_nan_fd,
            A_L_fd,
            yerr=abs(A_L_fd2 - A_L_fd1) / 2,
            markersize=10,
            fmt=marker10[2],
            color=color2[ii],
            mfc="white",
        )
        axis[i, j].errorbar(
            rN_nan_fd,
            A_rs_fd,
            yerr=abs(A_rs_fd2 - A_rs_fd1) / 2,
            markersize=10,
            fmt=marker10[4],
            color=color2[ii],
            mfc="white",
        )
        axis[i, j].errorbar(
            rN_nan_fd,
            A_CF_fd,
            yerr=abs(A_CF_fd2 - A_CF_fd1) / 2,
            markersize=10,
            fmt=marker10[3],
            color=color2[ii],
            mfc="white",
        )

        axis[i, j].scatter(
            rN_nan_fd,
            A_M_fd,
            s=160,
            marker=marker10[0],
            color=color2[ii],
            edgecolors="y",
            label="-[(" + lb[rp.name[1]] + ")" + " - " + "(" + lb[rp.name[0]] + ")]",
        )
        if j == 0:
            axis[i, j].legend(fontsize="14", loc="lower right")
        elif j == 1:
            axis[i, j].legend(fontsize="14", loc="upper right")

        ####
        i = 1
        if ii < len(r_positive) // 2:
            j = 0
        else:
            j = 1

        if ii == 4:  # ii == 0 or
            pp1 = axis[i, j].scatter(
                rN_nan_fd,
                -CRE_M_fd,
                s=140,
                marker=marker10[0],
                facecolors="none",
                edgecolors=color2[ii],
            )
            pp2 = axis[i, j].errorbar(
                rN_nan_fd,
                -CRE_T_fd,
                yerr=abs(CRE_T_fd2 - CRE_T_fd1) / 2,
                markersize=10,
                fmt=marker10[1],
                color=color2[ii],
                mfc="white",
            )
            pp3 = axis[i, j].errorbar(
                rN_nan_fd,
                -CRE_L_fd,
                yerr=abs(CRE_L_fd2 - CRE_L_fd1) / 2,
                markersize=10,
                fmt=marker10[2],
                color=color2[ii],
                mfc="white",
            )
            pp4 = axis[i, j].errorbar(
                rN_nan_fd,
                -CRE_CF_fd,
                yerr=abs(CRE_CF_fd2 - CRE_CF_fd1) / 2,
                markersize=10,
                fmt=marker10[3],
                color=color2[ii],
                mfc="white",
            )
            pp5 = axis[i, j].errorbar(
                rN_nan_fd,
                -CRE_rs_fd,
                yerr=abs(CRE_rs_fd2 - CRE_rs_fd1) / 2,
                markersize=10,
                fmt=marker10[4],
                color=color2[ii],
                mfc="white",
            )
        axis[i, j].errorbar(
            rN_nan_fd,
            -CRE_T_fd,
            yerr=abs(CRE_T_fd2 - CRE_T_fd1) / 2,
            markersize=10,
            fmt=marker10[1],
            color=color2[ii],
            mfc="white",
        )
        axis[i, j].errorbar(
            rN_nan_fd,
            -CRE_L_fd,
            yerr=abs(CRE_L_fd2 - CRE_L_fd1) / 2,
            markersize=10,
            fmt=marker10[2],
            color=color2[ii],
            mfc="white",
        )
        axis[i, j].errorbar(
            rN_nan_fd,
            -CRE_CF_fd,
            yerr=abs(CRE_CF_fd2 - CRE_CF_fd1) / 2,
            markersize=10,
            fmt=marker10[3],
            color=color2[ii],
            mfc="white",
        )
        axis[i, j].errorbar(
            rN_nan_fd,
            -CRE_rs_fd,
            yerr=abs(CRE_rs_fd2 - CRE_rs_fd1) / 2,
            markersize=10,
            fmt=marker10[4],
            color=color2[ii],
            mfc="white",
        )

        axis[i, j].scatter(
            rN_nan_fd,
            -CRE_M_fd,
            s=160,
            marker=marker10[0],
            color=color2[ii],
            edgecolors="y",
            label="-[(" + lb[rp.name[1]] + ")" + " - " + "(" + lb[rp.name[0]] + ")]",
        )

    leg2 = axis[1, 1].legend(
        [pp1, pp2, pp3, pp4, pp5],
        ["Total", "$N_c$", "LWP", "CF", "Residual"],
        fontsize="14",
        loc="upper left",
    )

    plt.tight_layout()  # w_pad=1
    return fig



import xarray as xr


class Simulation:
    def __init__(
        self, CWP, CF, NCCN, albedo, albd_clear, SOLIN, sw_dwn_profile=None, inv_idx=None
    ):
        self.quantities = self.Quantities(
            CWP, CF, NCCN, albedo, albd_clear, SOLIN, sw_dwn_profile, inv_idx
        )
        self.calc = self.Calculations(self.quantities)

    class Calculations:
        def __init__(self, quantities):
            self.q = quantities

        def cloud_albedo(self):
            # Calculate the cloud albedo (A2)
            self.q.cloud_albedo = (
                self.q.albedo - (1 - self.q.CF) * self.q.albd_clear
            ) / self.q.CF

        def free_trop_transmissivity(self):
            if self.q.inv_idx is not None:
                self.q.ft_transmissivity = self.q.sw_dwn_profile / self.q.SOLIN
            else:
                self.q.ft_transmissivity = 0.8

        def scene_albedo(self, a_ft=0.05):
            """a_ft = albedo free troposphere."""
            if self.q.ft_transmissivity is None:
                self.free_trop_transmissivity()
            T_ft = self.q.ft_transmissivity
            self.q.scene_albedo = (self.q.cloud_albedo - a_ft) / (
                T_ft**2 + self.q.cloud_albedo * a_ft - a_ft**2
            )

        def cre(self, level="cloud_layer", a_ft=0.05):
            """Calculation of cloud radiative effect (CRE)

            level : str
                cloud_layer: remove free-tropospheric effect with simple approx.
                toa: cloud radiative effect measured at top-of-atmosphere
            """
            if level == "cloud_layer":
                if self.q.scene_albedo is None:
                    self.scene_albedo(a_ft=a_ft)
                cre = -self.q.scene_albedo * self.q.SOLIN * self.q.CF
            elif level == "toa":
                cre = (
                    -1
                    * (self.q.cloud_albedo - self.q.albd_clear)
                    * self.q.SOLIN
                    * (self.q.CF)
                )
            return cre

    class Quantities:
        def __init__(
            self, CWP, CF, NCCN, albedo, albd_clear, SOLIN, sw_dwn_profile, inv_idx
        ):
            self.CWP = CWP
            self.CF = CF
            self.albedo = albedo
            self.albd_clear = albd_clear
            self.NCCN = NCCN
            self.SOLIN = SOLIN
            self.sw_dwn_profile = sw_dwn_profile
            self.inv_idx = inv_idx
            self.cloud_albedo = None
            self.ft_transmissivity = None
            self.scene_albedo = None
            self.cre = None


class CRE_Decompositer:
    def __init__(self, simulation1, simulation2):
        self.simulation1 = simulation1
        self.simulation2 = simulation2
        self.cre_scaling = True
        self.cloud_cre_change = None

    def cre_change(self, factor, cre_scaling=False):
        if self.simulation1.quantities.cre is None:
            self.simulation1.quantities.cre = self.simulation1.calc.cre()
        if self.simulation2.quantities.cre is None:
            self.simulation2.quantities.cre = self.simulation2.calc.cre()
        if cre_scaling:
            cre_direct = self.simulation1.calc.cre(level="toa")
            scale = cre_direct / self.simulation1.quantities.cre
            self.cre_scaling = scale
            self.simulation1.quantities.cre *= scale
            self.simulation2.quantities.cre *= scale
        self.CRE_change = factor * (
            self.simulation2.quantities.cre - self.simulation1.quantities.cre
        )

    def albedo_change(self):
        self.alb_change = (
            self.simulation2.quantities.cloud_albedo
            - self.simulation1.quantities.cloud_albedo
        )

    def cre_cloud_change(self, factor=-1):
        """CRE change due to changes in cloud properties.

        This is a more direct way to calculate the CRE change due to
        changes in cloud properties and should ideally match the sum of
        CRE changes due to LWP adjustment and twomey effect.
        """
        s1 = self.simulation1.quantities
        s2 = self.simulation2.quantities
        if self.cloud_cre_change is None:
            self.albedo_change()
        self.cloud_cre_change = factor * self.alb_change * s2.SOLIN * s1.CF

    def _integral(self, factor, albedo, x, sign=-1):
        a = albedo + sign * albedo * (1 - albedo) * (x ** (factor) - 1) / (
            1 + albedo * (x**factor - 1)
        )
        return a

    def _single_layer_model(self, Acld, alb, T):
        alpha = alb + (Acld * T**2) / (1 - alb * Acld)
        return alpha

    def twomey_effect(self, factor=-1, a_ft=0.05):
        s1 = self.simulation1.quantities
        s2 = self.simulation2.quantities
        rN = s2.NCCN / s1.NCCN  # .max(dim='height') / s1.NCCN.max(dim='height')
        self.twomey_rN = rN
        self.twomey_cld_albedo_change = self._integral(
            factor=1 / 3, albedo=s1.scene_albedo, x=rN, sign=factor
        )
        self.twomey_scene_albedo_change = self._single_layer_model(
            alb=a_ft, Acld=self.twomey_cld_albedo_change, T=s1.ft_transmissivity
        )
        self.twomey_cre_change = (
            -(self.twomey_scene_albedo_change - s1.cloud_albedo)
            * s2.SOLIN
            * s1.CF
            * self.cre_scaling
        )

    def LWP_adjustment(self, factor=-1, a_ft=0.05):
        s1 = self.simulation1.quantities
        s2 = self.simulation2.quantities
        rL = s2.CWP / s1.CWP
        self.LWP_rL = rL
        self.LWP_cld_albedo_change = self._integral(
            factor=5 / 6, albedo=s1.scene_albedo, x=rL, sign=factor
        )
        self.LWP_scene_albedo_change = self._single_layer_model(
            alb=a_ft, Acld=self.LWP_cld_albedo_change, T=s1.ft_transmissivity
        )
        self.LWP_cre_change = (
            -(self.LWP_scene_albedo_change - s1.cloud_albedo)
            * s2.SOLIN
            * s1.CF
            * self.cre_scaling
        )

    def CF_adjustment(self, factor=-1):
        s1 = self.simulation1.quantities
        s2 = self.simulation2.quantities
        A_CF = self.twomey_scene_albedo_change.copy()
        for i in range(len(s1.CF)):
            if (
                s2.CF[i] > s1.CF[i]
            ):  # what is the reasoning behind this? depending on the run with the higher cloud fraction, the albedo change to its reference is used. Works if clear-sky fluxes are available in cloudy columns.
                A_CF[i] = s1.cloud_albedo[i] + (s2.CF[i] - s1.CF[i]) * (
                    s2.cloud_albedo[i] - s2.albd_clear[i]
                )
            else:
                A_CF[i] = s1.cloud_albedo[i] + (s2.CF[i] - s1.CF[i]) * (
                    s1.cloud_albedo[i] - s1.albd_clear[i]
                )  # A1 gets added

        self.CF_albedo_change = A_CF
        self.CF_cre_change = (
            factor * (A_CF - s1.cloud_albedo) * s2.SOLIN * -1 * self.cre_scaling
        )

    def decompose(self, factor=-1, a_ft=0.05, cre_scaling=False):
        self.simulation1.calc.cloud_albedo()
        self.simulation2.calc.cloud_albedo()
        self.simulation1.calc.scene_albedo()
        self.simulation2.calc.scene_albedo()
        self.simulation1.quantities.cre = self.simulation1.calc.cre()
        self.simulation2.quantities.cre = self.simulation2.calc.cre()
        self.cre_change(factor=factor, cre_scaling=cre_scaling)
        self.albedo_change()
        self.cre_cloud_change(factor=factor)
        self.twomey_effect(factor=factor, a_ft=a_ft)
        self.LWP_adjustment(factor=factor, a_ft=a_ft)
        self.CF_adjustment(factor=factor)

        self.A_rs = (
            self.twomey_scene_albedo_change
            + self.LWP_scene_albedo_change
            + self.CF_albedo_change
            - self.alb_change
        )
        self.CRE_rs = (
            self.twomey_cre_change
            + self.LWP_cre_change
            + self.CF_cre_change
            - self.CRE_change
        )

        self.rC = self.simulation2.quantities.CF / self.simulation1.quantities.CF

    def to_dataset(self):
        """Return decomposition as dataset."""
        ds_cre = xr.Dataset(
            {
                "dCRE_total": self.CRE_change,
                "dCRE_CF": self.CF_cre_change,
                "dCRE_cloud": self.cloud_cre_change,
                "dCRE_CDNC": self.twomey_cre_change,
                "dCRE_LWP": self.LWP_cre_change,
                "dCRE_rs": self.CRE_rs,
            }
        )
        ds_alb = xr.Dataset(
            {
                "dA_total": self.alb_change,
                "dA_CF": self.CF_albedo_change,
                "dA_CDNC": self.twomey_scene_albedo_change,
                "dA_LWP": self.LWP_scene_albedo_change,
                "dA_rs": self.A_rs,
            }
        )
        ds_sim1 = xr.Dataset({"A_cloud_sim1": self.simulation1.quantities.cloud_albedo})
        ds_sim2 = xr.Dataset({"A_cloud_sim2": self.simulation2.quantities.cloud_albedo})
        ds = xr.merge([ds_cre, ds_alb, ds_sim1, ds_sim2])
        return ds

    def return_erfani_2022(self, otype="tuple"):
        """Matching return of quant_CRE_alltime."""
        if otype == "tuple":
            return (
                self.twomey_rN,
                self.LWP_rL,
                self.rC,
                self.CRE_change,
                self.twomey_cre_change,
                self.LWP_cre_change,
                self.CF_cre_change,
                self.CRE_rs,
                self.simulation1.quantities.cloud_albedo,
                self.twomey_scene_albedo_change,
                self.LWP_scene_albedo_change,
                self.CF_albedo_change,
                self.A_rs,
                self.simulation1.quantities.cre,
                self.simulation2.quantities.cre,
                self.simulation1.quantities.CF,
            )
        elif otype == "dict":
            return {
                "rN": self.twomey_rN,
                "rL": self.LWP_rL,
                "rC": self.rC,
                "CRE_M": self.CRE_change,
                "CRE_T": self.twomey_cre_change,
                "CRE_L": self.LWP_cre_change,
                "CRE_CF": self.CF_cre_change,
                "CRE_rs": self.CRE_rs,
                "A1": self.simulation1.quantities.cloud_albedo,
                "A_T": self.twomey_scene_albedo_change,
                "A_L": self.LWP_scene_albedo_change,
                "A_CF": self.CF_albedo_change,
                "A_rs": self.A_rs,
                "CRE1": self.simulation1.quantities.cre,
                "CRE2": self.simulation2.quantities.cre,
                "CF1": self.simulation1.quantities.CF,
            }



import numpy as np  # linear algebra

import dcre.erfani_2022.helpers as h  # noqa: E402


def quant_CRE_alltime_old(fn1, fn2, a_ft, box, factor=1, cre_scaling=False):
    (
        time,
        z,
        dz,
        RHO,
        CWP,
        RWP,
        ISCCPTOT,
        ZINV,
        entr,
        LWNTOA,
        SWNS,
        P,
        TABS,
        inv_idx,
        W2_Zinv_200,
        acc_prec,
        THETA,
        QT,
        integ_NA_BL,
        integ_NC_BL,
        COD,
        SAM_re,
        height_NARC,
        height_ENTRNMT,
        height_NASURF,
        height_NPRA,
        height_NASED,
        height_NASCAV,
        sfc_wnd,
        albedo,
        NA,
        NC,
        PREC,
        RH,
        albd_clear,
        SOLIN2,
        SW_CRE,
        SWNTOAC,
        PRECIP,
        CLD,
        W2,
        WOBS,
        QV,
    ) = h.read_model_vars(fn1)

    # CF1  = smooth(ISCCPTOT, 10)
    CF1 = ISCCPTOT
    #     T_ft1= SWNTOAC / SOLIN2  # old method: transmissivity of the free troposphere
    RADSWDN = fn1.variables["RADSWDN"][:]  # Downward shortwave radiative flux: W/m2
    RADSWDN_Zinv = SOLIN2.copy()
    RADSWDN_Zinv[:] = np.nan
    for ij in range(len(RADSWDN_Zinv)):
        RADSWDN_Zinv[ij] = RADSWDN[ij, inv_idx[ij] + 5]
    T_ft1 = RADSWDN_Zinv / SOLIN2  # new method: transmissivity of the free troposphere

    A1 = (albedo - (1 - CF1) * albd_clear) / CF1  # Chen B2
    # A1   = albedo
    a1 = (a_ft - A1) / (a_ft**2 - A1 * a_ft - T_ft1**2)
    # a1   = albedo #/ CF1
    N1 = np.concatenate([integ_NC_BL[:3], h.smooth(integ_NC_BL, box)[3:]])

    # L1   = h.smooth((CWP + RWP)/CF1, 10)
    L1 = h.smooth(CWP + RWP, 10)
    CRE1 = -a1 * SOLIN2 * CF1

    if cre_scaling is True:
        fact = SW_CRE / CRE1
        CRE1 = CRE1 * fact
    else:
        fact = 1
    # CRE1 = SW_CRE
    A1clr = albd_clear

    #####
    (
        _,
        _,
        _,
        _,
        CWP,
        RWP,
        ISCCPTOT,
        _,
        _,
        _,
        _,
        _,
        _,
        inv_idx,
        _,
        _,
        _,
        _,
        _,
        integ_NC_BL,
        _,
        _,
        _,
        _,
        _,
        _,
        _,
        _,
        _,
        albedo,
        _,
        _,
        _,
        _,
        albd_clear,
        SOLIN2,
        SW_CRE,
        _,
        _,
        _,
        _,
        _,
        _,
    ) = h.read_model_vars(fn2)

    CF2 = ISCCPTOT
    RADSWDN = fn2.variables["RADSWDN"][:]  # Downward shortwave radiative flux: W/m2
    RADSWDN_Zinv = SOLIN2.copy()
    RADSWDN_Zinv[:] = np.nan
    for ij in range(len(RADSWDN_Zinv)):
        RADSWDN_Zinv[ij] = RADSWDN[ij, inv_idx[ij] + 5]
    T_ft2 = RADSWDN_Zinv / SOLIN2  # new method: transmissivity of the free troposphere

    A2 = (albedo - (1 - CF2) * albd_clear) / CF2  # Chen B2 --> A2 = cloud albedo
    # A2   = albedo
    a2 = (a_ft - A2) / (a_ft**2 - A2 * a_ft - T_ft2**2)  # Chen B7 multiplied by -1
    N2 = np.concatenate([integ_NC_BL[:3], h.smooth(integ_NC_BL, box)[3:]])
    # L2   = h.smooth((CWP + RWP)/CF2, 10)
    L2 = h.smooth(CWP + RWP, 10)
    CRE2 = -a2 * SOLIN2 * CF2
    CRE2 = CRE2 * fact  # SW_CRE / CRE1 # Erfani et al. (2022)
    A2clr = albd_clear

    ###
    # CRE_M = - (A2 - A1) * SOLIN2 * fact * CF1
    CRE_M = factor * (CRE2 - CRE1)
    A_M = A2 - A1

    ### Twomey Effect:
    rN = N2 / N1
    a_T = a1 + factor * a1 * (1 - a1) * (
        rN ** (1 / 3) - 1
    ) / (  # complicated form of B6a derived from Wood (2021) Eq. 2
        1 + a1 * (rN ** (1 / 3) - 1)
    )
    A_T = a_ft + a_T * T_ft1**2 / (1 - a_ft * a_T)  # Chen B8 for Nc
    CRE_T = (
        -(A_T - A1) * SOLIN2 * CF1
    ) * fact  # Chen eq after B8, part for Nc; why fact(SW_CRE / CRE1)?

    ### LWP adjustment Effect:
    rL = L2 / L1
    a_L = a1 + factor * a1 * (1 - a1) * (rL ** (5 / 6) - 1) / (
        1 + a1 * (rL ** (5 / 6) - 1)
    )
    A_L = a_ft + a_L * T_ft1**2 / (1 - a_ft * a_L)  # Chen B8 for LWP
    CRE_L = (
        -(A_L - A1) * SOLIN2 * CF1
    ) * fact  # Chen eq after B8, part for Nc; why fact(SW_CRE / CRE1)?

    ### CF adjustment Effect:
    #    A_CF   = (CF2 - CF1) * (A2 - A2clr)
    #    A_CF   = A1 + (CF2 - CF1) * (A2 - A2clr)

    A_CF = A_T.copy()

    for i in range(len(CF1)):
        if (
            CF2[i] > CF1[i]
        ):  # what is the reasoning behind this? depending on the run with the higher cloud fraction, the albedo change to its reference is used. Works if clear-sky fluxes are available in cloudy columns.
            A_CF[i] = A1[i] + (CF2[i] - CF1[i]) * (A2[i] - A2clr[i])
        else:
            A_CF[i] = A1[i] + (CF2[i] - CF1[i]) * (A1[i] - A1clr[i])  # A1 gets added

    CRE_CF = -1 * factor * (A_CF - A1) * SOLIN2 * fact  # A1 gets subtracted again

    ### residual:
    A_rs = A_T + A_L + A_CF - A_M
    CRE_rs = CRE_T + CRE_L + CRE_CF - CRE_M

    rC = CF2 / CF1

    return (
        rN,
        rL,
        rC,
        CRE_M,
        CRE_T,
        CRE_L,
        CRE_CF,
        CRE_rs,
        A1,
        A_T,
        A_L,
        A_CF,
        A_rs,
        CRE1,
        CRE2,
        CF1,
    )



#!/usr/bin/env python
# coding: utf-8
import argparse

import numpy as np  # linear algebra
import pandas as pd
from netCDF4 import Dataset

import dcre.erfani_2022.erfani_2022_results as er  # noqa: E402
import dcre.erfani_2022.figures as ef  # noqa: E402
import dcre.erfani_2022.helpers as h  # noqa: E402

lb = [
    "40-40-LD",
    "40-40",
    "40-40to150",
    "150-40",
    "40-150",
    "MERRA-LD",
    "MERRA",
    "MERRAX3",
    "250-60-LD",
    "250-60",
    "70-60-LD",
    "70-60",
    "110-60",
    "250-200",
    "MERRA",
    "MERRAX3",
]

ylim = 4000
box = 5

# Read sample file for time/mask information
sample_file = "./data/OUT_STAT/CSET_RF06_Tr2p3_256sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_40_40.nc"
(
    time,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    SWNS,
    P,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
    _,
) = h.read_model_vars(Dataset(sample_file))

upbound = time.copy()
upbound[:] = ylim
upbound[SWNS > 0] = np.nan


def plot_figures(fn):
    """Wrapper function for plotting CRE decomposition figures following Erfani
    et al. (2022)

    Inputs
    ------
    fn : str
        filename of pickle file containing CRE decomposition
    """
    res = pd.read_pickle(fn)

    # Fig 13
    j1 = [0, 0, 0, 0, 8, 8, 8, 8]
    j2 = [5, 3, 4, 7, 9, 13, 14, 15]
    it = "2015-07-17 00Z"
    it2 = "2015-07-27 00Z"
    init_time_all = [it, it, it, it, it2, it2, it2, it2]

    res_sub = er.select_res_subset(res, j1, j2)

    fig13 = ef.fig13(res_sub, lb)
    fig13.savefig("tests/decomp_CRE_fig13.pdf", format="pdf", dpi=500)

    # Fig 14
    fig14 = ef.fig14(res_sub, lb)
    fig14.savefig("tests/decomp_ratio_fig14.pdf", format="pdf", dpi=500)

    # Fig timeseries

    j1 = [6, 1, 1, 1, 14, 12, 12, 9]
    j2 = [7, 3, 4, 7, 15, 9, 15, 15]

    res_sub = er.select_res_subset(res, j1, j2)
    fig = ef.cre_timeseries(res_sub, time, lb, upbound, init_time_all)
    fig.savefig("tests/decomp_CRE_timeseries.pdf", format="pdf", dpi=500)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Plot CRE decomposition figures following Erfani et al. (2022)."
    )
    parser.add_argument(
        "-i",
        "--input",
        metavar="i",
        type=str,
        help="input pickle file containing CRE components",
    )
    args = parser.parse_args()
    plot_figures(args.input)



import matplotlib.pyplot as plt
import numpy as np

import dcre.erfani_2022.helpers as h  # noqa: E402


def calc_strong_tseries(var, var2, perc):
    """Calculate the timeseries of var2 when averaged over the 5-10% of the
    domain with the largest var."""
    outx = var[:, 0, 0].copy()
    outx[:] = np.nan
    for i in range(var.shape[0]):
        var_x = np.percentile(var[i, :, :], perc)
        ix = np.where(var[i, :, :] >= var_x)
        outx[i] = np.nanmean(var2[i, :, :][ix])
    return outx


def smooth_ends(var, box):
    out = var.copy()
    out = h.smooth(var, box)
    out[:box] = np.nan
    out[-box:] = np.nan
    # out = np.concatenate([var[:box], smooth(var, box)[box:]])#-box, var[-box:]])
    return out


def transition_timeseries(fn3d, cases, box, styles):
    PANELS = ["a)", "b)", "c)", "d)"]
    lb = styles["labels"]
    ls = styles["linestyles"]
    lw = styles["linewidths"]
    clr = styles["label_group_colors"]

    fig1, axis = plt.subplots(4, 1, figsize=(14, 18))

    i = 0
    axis[i].set_ylabel("LCC (%)", fontsize=22)
    axis[i].set_ylim([0, 105])
    axis[i].set_xlim([-1, 1])
    axis[i].axes.get_xaxis().set_ticklabels([])
    axis[i].tick_params(axis="both", which="major", labelsize=18)
    axis[i].grid(linestyle=":", linewidth=1)
    axis[i].text(
        0.04,
        1.1,
        PANELS[i],
        transform=axis[i].transAxes,
        fontsize=22,
        fontweight="bold",
        va="top",
        fontfamily="serif",
        ha="right",
    )

    i = 1
    axis[i].set_ylabel("LWP ($g$ $m^{-2}$)", fontsize=22)
    axis[i].set_ylim([0, 170])
    axis[i].set_xlim([-1, 1])
    axis[i].axes.get_xaxis().set_ticklabels([])
    axis[i].tick_params(axis="both", which="major", labelsize=18)
    axis[i].grid(linestyle=":", linewidth=1)
    axis[i].text(
        0.04,
        1.1,
        PANELS[i],
        transform=axis[i].transAxes,
        fontsize=22,
        fontweight="bold",
        va="top",
        fontfamily="serif",
        ha="right",
    )

    i = 2
    axis[i].set_ylabel("95$^{th}$ pctl LWP $<N_c>^{-1}$", fontsize=22)
    axis[i].set_ylim([0, 25])
    axis[i].set_xlim([-1, 1])
    axis[i].axes.get_xaxis().set_ticklabels([])
    axis[i].tick_params(axis="both", which="major", labelsize=18)
    axis[i].grid(linestyle=":", linewidth=1)
    axis[i].text(
        0.04,
        1.1,
        PANELS[i],
        transform=axis[i].transAxes,
        fontsize=22,
        fontweight="bold",
        va="top",
        fontfamily="serif",
        ha="right",
    )

    i = 3
    axis[i].set_xlabel("Time (SCT as 0)", fontsize=22)
    axis[i].set_ylabel("95$^{th}$ pctl Prec. ($mm$ $d^{-1}$)", fontsize=22)
    axis[i].set_xlim([-1, 1])
    axis[i].tick_params(axis="both", which="major", labelsize=18)
    axis[i].grid(linestyle=":", linewidth=1)
    axis[i].text(
        0.04,
        1.1,
        PANELS[i],
        transform=axis[i].transAxes,
        fontsize=22,
        fontweight="bold",
        va="top",
        fontfamily="serif",
        ha="right",
    )

    ####
    for jj in cases:
        T3D = fn3d[jj].variables["time"][:]  # time: day
        t3d = T3D - int(T3D[0])
        #    x        = fn3d[jj].variables['x'][:] / 1000   # $km$
        #    y        = fn3d[jj].variables['y'][:] / 1000   # $km$
        CLD2d = fn3d[jj].variables["CLD"][:]
        Prec = fn3d[jj].variables["Prec"][:]  # * 1000    # Surface Precip. Rate: g/m
        QCPATH = fn3d[jj].variables["QCPATH"][
            :
        ]  # * 1000    # Column cloud liquid mass (vertically integrated), before conversion to unit: g/m2
        NCQCPATH = fn3d[jj].variables["NCQCPATH"][
            :
        ]  # Column integral of QC*NC (useful for computing mass-weighted NC): kg/kg/m2
        CDNC = NCQCPATH / QCPATH  # #/mg
        CDNC = CDNC * 1e-6  # #/mg
        #    x_mg, y_mg = np.meshgrid(x, y) # making meshgrid

        CWP3d = QCPATH * 1000  # + QP
        CWP_Nc = CWP3d / CDNC
        #   looking at the timeseries of CWP/Nd when averaged over the 5-10% of the domain with the largest CWP.
        CWP_Nc95 = calc_strong_tseries(CWP3d, CWP_Nc, 95)
        Prec95 = calc_strong_tseries(Prec, Prec, 95)
        #    Prec95 = calc_strong_tseries(Prec, Prec, 95)
        #    CWP3d95 = calc_strong_tseries(CWP3d, CWP3d, 95)

        CLD2d_m = np.nanmean(np.nanmean(CLD2d, 2), 1)
        i_50 = np.where((CLD2d_m > 48) & ((CLD2d_m < 52)))[0]
        if len(i_50) == 0:
            i_50 = np.where((CLD2d_m > 45) & ((CLD2d_m < 55)))[0]
        if lb[jj] == "110-60" or lb[jj] == "250-60-LD":
            i_sct = -1
        else:
            i_sct = 0
        t_sct = t3d[i_50[i_sct]]

        t_new = t3d - t_sct

        i = 0
        axis[i].plot(
            t_new,
            smooth_ends(CLD2d_m, box),
            ls[jj],
            linewidth=lw[jj],
            color=clr[jj],
            label=lb[jj],
        )
        i = 1
        axis[i].plot(
            t_new,
            smooth_ends(np.nanmean(np.nanmean(CWP3d, 2), 1), box),
            ls[jj],
            linewidth=lw[jj],
            color=clr[jj],
            label=lb[jj],
        )
        i = 2
        axis[i].plot(
            t_new,
            smooth_ends(CWP_Nc95, box),
            ls[jj],
            linewidth=lw[jj],
            color=clr[jj],
            label=lb[jj],
        )
        i = 3
        axis[i].plot(
            t_new,
            smooth_ends(Prec95, box),
            ls[jj],
            linewidth=lw[jj],
            color=clr[jj],
            label=lb[jj],
        )
        axis[i].legend(fontsize=17.5, loc="upper left", ncol=2)
    plt.tight_layout()  # w_pad=1
    plt.savefig("all_SCT.pdf", format="pdf", dpi=500)
    return fig1



import numpy as np  # linear algebra

from . import erfani_output_decomp_new as dn
from . import erfani_output_decomp_orig as do


def quant_CRE_alltime(
    fn1, fn2, a_ft, box, factor=1, cre_scaling=False, mode="erfani_2022"
):
    if mode == "erfani_2022":
        return do.quant_CRE_alltime_old(
            fn1, fn2, a_ft, box, factor=factor, cre_scaling=cre_scaling
        )
    elif mode == "decomposer":
        return dn.quant_CRE_alltime_new(
            fn1, fn2, a_ft, box, factor=factor, cre_scaling=cre_scaling
        )


def quant_CRE_avg(j1j2, j2j1, output="normal"):
    (
        rN1,
        rL1,
        rC1,
        CRE_M1,
        CRE_T1,
        CRE_L1,
        CRE_CF1,
        CRE_rs1,
        A1,
        A_T1,
        A_L1,
        A_CF1,
        A_rs1,
        CRE1,
        CRE2,
        CF1,
    ) = j1j2
    (
        rN2,
        rL2,
        rC2,
        CRE_M2,
        CRE_T2,
        CRE_L2,
        CRE_CF2,
        CRE_rs2,
        A1_2,
        A_T2,
        A_L2,
        A_CF2,
        A_rs2,
        CRE1_2,
        CRE2_2,
        CF2,
    ) = j2j1

    rN = (rN1 + 1 / rN2) / 2
    rL = (rL1 + 1 / rL2) / 2
    rC = (rC1 + 1 / rC2) / 2

    CRE_T = (CRE_T1 + CRE_T2) / 2
    CRE_L = (CRE_L1 + CRE_L2) / 2
    CRE_CF = (CRE_CF1 + CRE_CF2) / 2
    CRE_rs = (CRE_rs1 + CRE_rs2) / 2
    CRE_M = (CRE_M1 + CRE_M2) / 2

    A_T = (A_T1 - A1 + A_T2 - A1_2) / 2
    A_L = (A_L1 - A1 + A_L2 - A1_2) / 2
    A_CF = (A_CF1 - A1 + A_CF2 - A1_2) / 2
    A_rs = (A_rs1 - A1 + A_rs2 - A1_2) / 2
    A_M = A1_2 - A1

    CF = (CF1 + CF2) / 2

    CF1_nan = CF.copy()  # CF1.copy()?
    CF2_nan = CF.copy()  # CF2.copy()?
    CF_nan = CF.copy()
    # CF1_nan[np.isnan(SOLIN2) == 1] = np.nan
    # CF2_nan[np.isnan(SOLIN2) == 1] = np.nan
    # CF_nan[np.isnan(SOLIN2) == 1] = np.nan

    CF1_nan_fd = np.nanmean(CF1_nan[10:-10])
    CF2_nan_fd = np.nanmean(CF2_nan[10:-10])
    CF_nan_fd = np.nanmean(CF_nan[10:-10])

    rN_nan = rN.copy()
    rL_nan = rL.copy()
    rC_nan = rC.copy()
    # rN_nan[np.isnan(SOLIN2) == 1] = np.nan
    # rL_nan[np.isnan(SOLIN2) == 1] = np.nan
    # rC_nan[np.isnan(SOLIN2) == 1] = np.nan

    rN_nan_fd = np.nanmean(rN_nan[10:-10])
    rL_nan_fd = np.nanmean(rL_nan[10:-10])
    rC_nan_fd = np.nanmean(rC_nan[10:-10])

    CRE_M_fd = np.nanmean(CRE_M[10:-10])
    CRE_T_fd = np.nanmean(CRE_T[10:-10])
    CRE_L_fd = np.nanmean(CRE_L[10:-10])
    CRE_CF_fd = np.nanmean(CRE_CF[10:-10])
    CRE_rs_fd = np.nanmean(CRE_rs[10:-10])

    A_M_fd = np.nanmean(A_M[10:-10])
    A_T_fd = np.nanmean(A_T[10:-10])
    A_L_fd = np.nanmean(A_L[10:-10])
    A_CF_fd = np.nanmean(A_CF[10:-10])
    A_rs_fd = np.nanmean(A_rs[10:-10])

    if output == "extended":
        CRE_M_fd1 = np.nanmean(CRE_M1[10:-10])
        CRE_T_fd1 = np.nanmean(CRE_T1[10:-10])
        CRE_L_fd1 = np.nanmean(CRE_L1[10:-10])
        CRE_CF_fd1 = np.nanmean(CRE_CF1[10:-10])
        CRE_rs_fd1 = np.nanmean(CRE_rs1[10:-10])

        CRE_M_fd2 = np.nanmean(CRE_M2[10:-10])
        CRE_T_fd2 = np.nanmean(CRE_T2[10:-10])
        CRE_L_fd2 = np.nanmean(CRE_L2[10:-10])
        CRE_CF_fd2 = np.nanmean(CRE_CF2[10:-10])
        CRE_rs_fd2 = np.nanmean(CRE_rs2[10:-10])

        A_M_fd1 = np.nanmean(A1[10:-10])
        A_T_fd1 = np.nanmean((A_T1 - A1)[10:-10])
        A_L_fd1 = np.nanmean((A_L1 - A1)[10:-10])
        A_CF_fd1 = np.nanmean((A_CF1 - A1)[10:-10])
        A_rs_fd1 = np.nanmean((A_rs1 - A1)[10:-10])

        A_M_fd2 = np.nanmean(A1_2[10:-10])
        A_T_fd2 = np.nanmean((A_T2 - A1_2)[10:-10])
        A_L_fd2 = np.nanmean((A_L2 - A1_2)[10:-10])
        A_CF_fd2 = np.nanmean((A_CF2 - A1_2)[10:-10])
        A_rs_fd2 = np.nanmean((A_rs2 - A1_2)[10:-10])

    if output == "normal":
        return (
            rN_nan_fd,
            rL_nan_fd,
            rC_nan_fd,
            CRE_M_fd,
            CRE_T_fd,
            CRE_L_fd,
            CRE_CF_fd,
            CRE_rs_fd,
            A_M_fd,
            A_T_fd,
            A_L_fd,
            A_CF_fd,
            A_rs_fd,
            CF1_nan_fd,
            CF2_nan_fd,
            CF_nan_fd,
        )
    elif output == "extended":
        return (
            rN_nan_fd,
            rL_nan_fd,
            rC_nan_fd,
            CRE_M_fd,
            CRE_T_fd,
            CRE_L_fd,
            CRE_CF_fd,
            CRE_rs_fd,
            A_M_fd,
            A_T_fd,
            A_L_fd,
            A_CF_fd,
            A_rs_fd,
            CRE_M_fd1,
            CRE_T_fd1,
            CRE_L_fd1,
            CRE_CF_fd1,
            CRE_rs_fd1,
            A_M_fd1,
            A_T_fd1,
            A_L_fd1,
            A_CF_fd1,
            A_rs_fd1,
            CRE_M_fd2,
            CRE_T_fd2,
            CRE_L_fd2,
            CRE_CF_fd2,
            CRE_rs_fd2,
            A_M_fd2,
            A_T_fd2,
            A_L_fd2,
            A_CF_fd2,
            A_rs_fd2,
            CF1_nan_fd,
            CF2_nan_fd,
            CF_nan_fd,
        )


def quant_CRE(
    fn1, fn2, a_ft, box, cre_scaling=False, mode="erfani_2022", output="normal"
):
    j1j2 = quant_CRE_alltime(
        fn1, fn2, a_ft, box, factor=1, cre_scaling=cre_scaling, mode=mode
    )
    j2j1 = quant_CRE_alltime(
        fn2, fn1, a_ft, box, factor=-1, cre_scaling=cre_scaling, mode=mode
    )
    return quant_CRE_avg(j1j2, j2j1)



#!/usr/bin/env python
# coding: utf-8
import os
import warnings

import numpy as np  # linear algebra
import pandas as pd
import pytest

import dcre.erfani_2022.erfani_2022_results as er  # noqa: E402

warnings.filterwarnings("ignore")

cre_scaling = False
fn_res_orig = f"cre_components_erfani_scaling{cre_scaling}.pkl"
fn_res_new = f"cre_components_decomposer_scaling{cre_scaling}.pkl"
abs_rel_error_tol = 0.2  # [%]


# Calculate and decompose CRE values
def get_decomposition(
    mode, cre_scaling=False, a_ft=0.05, box=5, reuse=False, cache=None
):
    if reuse:
        assert cache is not None, "Cache must be given when reuse=True"
        if os.path.exists(cache):
            res = pd.read_pickle(cache)
        else:
            print("Data could not be found in cache. Recomputing.")
            res = er.calc_results(a_ft=a_ft, box=box, scaling=cre_scaling, mode=mode)
            res.to_pickle(cache)
    else:
        res = er.calc_results(a_ft=a_ft, box=box, scaling=cre_scaling, mode=mode)
        if cache:
            res.to_pickle(cache)
    return res


@pytest.mark.parametrize("scaling", [False, True])
def test_decomposition_against_erfani_2022(scaling, **kwargs):
    res = {}
    for mode in ["erfani_2022", "decomposer"]:
        c = f"tests/cre_components_{mode}_scaling{scaling}.pkl"
        res[mode] = get_decomposition(mode, cre_scaling=scaling, cache=c, **kwargs)

    for g, grp in pd.concat([res["erfani_2022"], res["decomposer"]]).groupby(
        level=[0, 1, 2, 3, 4, 5]
    ):
        assert (
            len(grp) == 2
        ), f"For the index {g} only one of the datasets seem to have an entry."
        rel_diff = (grp.iloc[0] - grp.iloc[1]) / grp.iloc[0] * 100
        abs_rel_diff = np.abs(rel_diff)
        max_rel_diff = abs_rel_diff.apply(np.max)
        print(max_rel_diff)
        assert np.all(
            max_rel_diff < abs_rel_error_tol
        ), "Maximum relative error exceeds threshold."



import matplotlib.pyplot as plt

from .. import erfani_decomp_func as ed  # noqa: E402

PANELS1 = ["a)", "b)", "c)", "d)", "e)", "f)", "g)", "h)"]
color2 = ["k", "b", "tab:red", "tab:green", "k", "b", "tab:red", "tab:green"]
marker = ["o", "x", "s", "*", "^", "+", "d", "p"]


def fig14(res, lb):
    r_positive = res.loc[
        (slice(None), slice(None), slice(None), slice(None), slice(None), 1)
    ]
    r_negative = res.loc[
        (slice(None), slice(None), slice(None), slice(None), slice(None), -1)
    ]
    assert len(r_positive) == len(r_negative), "Datasets must have same length."

    fig, axis = plt.subplots(2, 2, figsize=(11, 10))

    i = 0
    j = 0
    axis[i, j].set_ylabel("$r_C$", fontsize="18")
    axis[i, j].set_xlabel("$r_N$", fontsize="18")
    axis[i, j].set_title("L06", fontsize="23")
    axis[i, j].set_xlim([0.5, 10])
    axis[i, j].set_ylim([0.75, 3.5])
    axis[i, j].tick_params(axis="both", which="major", labelsize=14)
    axis[i, j].grid(linestyle=":", linewidth=1)  # axis='y',
    axis[i, j].text(
        0.1,
        1.08,
        PANELS1[i * 2 + j],
        transform=axis[i, j].transAxes,
        fontsize=18,
        fontweight="bold",
        va="top",
        fontfamily="serif",
        ha="right",
    )

    ####
    i = 0
    j = 1
    axis[i, j].set_xlabel("$r_N$", fontsize="18")
    axis[i, j].set_title("L10", fontsize="23")
    axis[i, j].set_xlim([0.5, 10])
    axis[i, j].set_ylim([0.75, 3.5])
    axis[i, j].tick_params(axis="both", which="major", labelsize=14)
    axis[i, j].grid(linestyle=":", linewidth=1)
    axis[i, j].text(
        0.1,
        1.08,
        PANELS1[i * 2 + j],
        transform=axis[i, j].transAxes,
        fontsize=18,
        fontweight="bold",
        va="top",
        fontfamily="serif",
        ha="right",
    )

    ####
    i = 1
    j = 0
    axis[i, j].set_ylabel("$r_N$", fontsize="18")
    axis[i, j].set_xlabel("$r_L$", fontsize="18")
    axis[i, j].set_xlim([0.8, 1.75])
    axis[i, j].set_ylim([0.5, 10])
    axis[i, j].tick_params(axis="both", which="major", labelsize=14)
    axis[i, j].grid(linestyle=":", linewidth=1)
    axis[i, j].text(
        0.1,
        1.08,
        PANELS1[i * 2 + j],
        transform=axis[i, j].transAxes,
        fontsize=18,
        fontweight="bold",
        va="top",
        fontfamily="serif",
        ha="right",
    )

    ####
    i = 1
    j = 1
    axis[i, j].set_xlabel("$r_L$", fontsize="18")
    axis[i, j].set_xlim([0.8, 1.75])
    axis[i, j].set_ylim([0.5, 10])
    axis[i, j].tick_params(axis="both", which="major", labelsize=14)
    axis[i, j].grid(linestyle=":", linewidth=1)
    axis[i, j].text(
        0.1,
        1.08,
        PANELS1[i * 2 + j],
        transform=axis[i, j].transAxes,
        fontsize=18,
        fontweight="bold",
        va="top",
        fontfamily="serif",
        ha="right",
    )

    #####
    for ii in range(len(r_positive)):
        rp = r_positive.iloc[ii]
        rn = r_negative.iloc[ii]

        (
            rN_nan_fd,
            rL_nan_fd,
            rC_nan_fd,
            CRE_M_fd,
            CRE_T_fd,
            CRE_L_fd,
            CRE_CF_fd,
            CRE_rs_fd,
            A_M_fd,
            A_T_fd,
            A_L_fd,
            A_CF_fd,
            A_rs_fd,
            CF1_nan_fd,
            CF2_nan_fd,
            CF_nan_fd,
        ) = ed.quant_CRE_avg(rp, rn, output="normal")

        ####
        if ii < len(r_positive) // 2:
            i = 0
            j = 0
        else:
            i = 0
            j = 1
        axis[i, j].scatter(
            rN_nan_fd,
            rC_nan_fd,
            s=160,
            marker=marker[0],
            color=color2[ii],
            label="(" + lb[rp.name[1]] + ")" + " / " + "(" + lb[rp.name[0]] + ")",
        )
        axis[i, j].legend(fontsize="14")

        ####
        if ii < len(r_positive) // 2:
            i = 1
            j = 0
        else:
            i = 1
            j = 1
        axis[i, j].scatter(
            rL_nan_fd,
            rN_nan_fd,
            s=160,
            marker=marker[0],
            color=color2[ii],
            label="(" + lb[rp.name[1]] + ")" + " / " + "(" + lb[rp.name[0]] + ")",
        )

    plt.tight_layout()
    return fig



#!/usr/bin/env python
# coding: utf-8

import hashlib

import pandas as pd
from netCDF4 import Dataset

from . import erfani_decomp_func as ed  # noqa: E402

##### read various files and variables
## model runs

datasets = {
    "fn2": [
        "./data/OUT_STAT/CSET_RF06_Tr2p3_256sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_40_40.nc",
        "./data/OUT_STAT/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_va_1_4.nc",
        "./data/OUT_STAT/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_BL40_FT40_150.nc",
        "./data/OUT_STAT/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_vc_BL_150.nc",
        "./data/OUT_STAT/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_vb_FTNA_150.nc",
        "./data/OUT_STAT/CSET_RF06_Tr2p3_256sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_MERRA2.nc",
        "./data/OUT_STAT/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_MERRA2loglogMBL.nc",
        "./data/OUT_STAT/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_MERRA2loglogMBL_3t.nc",
        "./data/OUT_STAT/CSET_RF10_Tr6p0_256sqx432_100m_RF10_M2005PA_RRTM4PBL_UM5_ProgAer_250_60.nc",
        "./data/OUT_STAT/CSET_RF10_Tr6p0_96x96x432_100m_M2005_RRTM4PBL_UM5_RF10_Tr6p0_va.nc",
        "./data/OUT_STAT/CSET_RF10_Tr6p0_256sqx432_100m_RF10_M2005PA_RRTM4PBL_UM5_ProgAer_70_60.nc",
        "./data/OUT_STAT/CSET_RF10_Tr6p0_96sqx432_100m_RF10_M2005PA_RRTM4PBL_UM5_ProgAer_70_60_NEW.nc",
        "./data/OUT_STAT/CSET_RF10_Tr6p0_96x96x432_100m_M2005_RRTM4PBL_UM5_RF10_Tr6p0_vc_BL_70.nc",
        "./data/OUT_STAT/CSET_RF10_Tr6p0_96x96x432_100m_M2005_RRTM4PBL_UM5_RF10_Tr6p0_vb_FTNA_200.nc",
        "./data/OUT_STAT/CSET_RF10_Tr6p0_96sqx432_100m_RF10_Tr6p0_M2005PA_RRTM4PBL_UM5_ProgAer_MERRA2loglogMBL.nc",
        "./data/OUT_STAT/CSET_RF10_Tr6p0_96sqx432_100m_Tr6p0_M2005PA_RRTM4PBL_UM5_MERRAlogBL_3t.nc",
    ],
    "fn3d": [
        "./data/OUT_2D/CSET_RF06_Tr2p3_256sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_40_40_128.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_va_48.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_BL40_FT40_150_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_vc_BL_150_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_vb_FTNA_150_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF06_Tr2p3_256sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_MERRA2_128.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_MERRA2loglogMBL_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_MERRA2loglogMBL_3t_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_256sqx432_100m_RF10_M2005PA_RRTM4PBL_UM5_ProgAer_250_60_128.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_96x96x432_100m_M2005_RRTM4PBL_UM5_RF10_Tr6p0_va_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_256sqx432_100m_RF10_M2005PA_RRTM4PBL_UM5_ProgAer_70_60_128.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_96sqx432_100m_RF10_M2005PA_RRTM4PBL_UM5_ProgAer_70_60_NEW_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_96x96x432_100m_M2005_RRTM4PBL_UM5_RF10_Tr6p0_vc_BL_70_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_96x96x432_100m_M2005_RRTM4PBL_UM5_RF10_Tr6p0_vb_FTNA_200_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_96sqx432_100m_RF10_Tr6p0_M2005PA_RRTM4PBL_UM5_ProgAer_MERRA2loglogMBL_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_96sqx432_100m_Tr6p0_M2005PA_RRTM4PBL_UM5_MERRAlogBL_3t_96.2Dbin_1.nc",
    ],
    "fn2d": [
        "./data/OUT_2D/CSET_RF06_Tr2p3_256sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_40_40_128.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_va_48.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_BL40_FT40_150_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_vc_BL_150_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_vb_FTNA_150_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF06_Tr2p3_256sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_MERRA2_128.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_MERRA2loglogMBL_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF06_Tr2p3_96sqx432_100m_M2005PA_RRTM4PBL_UM5_ProgAer_MERRA2loglogMBL_3t_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_256sqx432_100m_RF10_M2005PA_RRTM4PBL_UM5_ProgAer_250_60_128.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_96x96x432_100m_M2005_RRTM4PBL_UM5_RF10_Tr6p0_va_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_256sqx432_100m_RF10_M2005PA_RRTM4PBL_UM5_ProgAer_70_60_128.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_96sqx432_100m_RF10_M2005PA_RRTM4PBL_UM5_ProgAer_70_60_NEW_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_96x96x432_100m_M2005_RRTM4PBL_UM5_RF10_Tr6p0_vc_BL_70_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_96x96x432_100m_M2005_RRTM4PBL_UM5_RF10_Tr6p0_vb_FTNA_200_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_96sqx432_100m_RF10_Tr6p0_M2005PA_RRTM4PBL_UM5_ProgAer_MERRA2loglogMBL_96.2Dbin_1.nc",
        "./data/OUT_2D/CSET_RF10_Tr6p0_96sqx432_100m_Tr6p0_M2005PA_RRTM4PBL_UM5_MERRAlogBL_3t_96.2Dbin_1.nc",
    ],
}


def open_dataset(filelist):
    f = [Dataset(fn) for fn in filelist]
    return f


def add_combinations(
    filenames,
    j1,
    j2,
    a_ft,
    box,
    cre_scaling=False,
    factor=None,
    mode="erfani_2022",
    combinations=None,
):
    if combinations is None:
        combinations = {}

    def hash(obj):
        return hashlib.sha256(str(obj.__dict__).encode()).hexdigest()

    for i in range(len(j1)):
        fn1 = filenames[j1[i]]
        fn2 = filenames[j2[i]]
        res = get_dCRE(
            fn1, fn2, a_ft, box, cre_scaling=cre_scaling, factor=factor, mode=mode
        )
        combinations[j1[i], j2[i], a_ft, box, cre_scaling, factor, mode] = res
    return combinations


def get_dCRE(fn1, fn2, a_ft, box, cre_scaling=False, factor=1, mode="erfani_2022"):
    components = ed.quant_CRE_alltime(
        fn1, fn2, a_ft, box, cre_scaling=cre_scaling, factor=factor, mode=mode
    )
    return components


def calc_results(a_ft, box, scaling=False, mode="erfani_2022"):
    fn2 = open_dataset(datasets["fn2"])

    j1 = [0, 0, 0, 0, 8, 8, 8, 8]
    j2 = [5, 3, 4, 7, 9, 13, 14, 15]

    combinations = add_combinations(
        fn2, j1, j2, a_ft, box, cre_scaling=scaling, factor=1, mode=mode
    )
    combinations = add_combinations(
        fn2,
        j2,
        j1,
        a_ft,
        box,
        cre_scaling=scaling,
        factor=-1,
        mode=mode,
        combinations=combinations,
    )

    j1 = [6, 1, 1, 1, 14, 12, 12, 9]
    j2 = [7, 3, 4, 7, 15, 9, 15, 15]
    combinations = add_combinations(
        fn2,
        j1,
        j2,
        a_ft,
        box,
        cre_scaling=scaling,
        factor=1,
        mode=mode,
        combinations=combinations,
    )
    combinations = add_combinations(
        fn2,
        j2,
        j1,
        a_ft,
        box,
        cre_scaling=scaling,
        factor=-1,
        mode=mode,
        combinations=combinations,
    )

    df = pd.DataFrame.from_dict(combinations, orient="index")
    df.set_index(pd.MultiIndex.from_tuples(df.index), inplace=True)
    return df


def get_result(df, j1, j2, a_ft, box, cre_scaling, factor, mode):
    idx = (j1, j2, a_ft, box, cre_scaling, factor, mode)
    return tuple(df.loc[idx])


def select_res_subset(res, j1, j2, a_ft=None, box=None, cre_scaling=None):
    r_positive = res.loc[
        (slice(None), slice(None), slice(None), slice(None), slice(None), [1])
    ]
    r_negative = res.loc[
        (slice(None), slice(None), slice(None), slice(None), slice(None), [-1])
    ]
    r_negative = r_negative.swaplevel(0, 1)
    res_ = pd.concat([r_positive, r_negative])
    return res_.loc[
        res_.index.get_level_values(0).isin(j1) & res_.index.get_level_values(1).isin(j2)
    ]



import numpy as np


def read_model_vars(file):  # noqa: C901
    def height_Ave(var):
        integ_var = np.arange(var[:, 0].size).astype(float)
        integ_var[:] = np.nan
        for i in range(var[:, 0].size):
            integ_var[i] = np.nansum(
                dz[: inv_idx[i] + 5]
                * RHO[i, : inv_idx[i] + 5]
                * var[i, : inv_idx[i] + 5]
            ) / np.nansum(dz[: inv_idx[i] + 5] * RHO[i, : inv_idx[i] + 5])
        return integ_var

    ## model variables:
    TIME = file.variables["time"][:]
    # lat_mg, lon_mg = np.meshgrid(lat, lon) # making meshgrid
    # 1-D variables
    z = file.variables["z"][:]
    SST = (
        file.variables["SST"][:] - 273.15
    )  # sea surface temperature converted from kelvin to celcius unit
    SSTOBS = file.variables["SSTOBS"][:] - 273.15
    PREC = file.variables["PREC"][:]  # Surface Precipitation"  "mm/day"
    LHF = file.variables["LHF"][:]
    LHFOBS = file.variables["LHFOBS"][:]
    SHF = file.variables["SHF"][:]
    SHFOBS = file.variables["SHFOBS"][:]
    Ps = file.variables["Ps"][:]
    LWNS = file.variables["LWNS"][:]
    SWNS = file.variables["SWNS"][:]
    RWP = file.variables["RWP"][:]  # Rain Water Path: g/m2
    CWP = file.variables["CWP"][:]  # Cloud Water Path
    LWP = file.variables["LWP"][:]  # GCSS Liquid Water Path
    MODISLWP = file.variables["MODISLWP"][:]
    CLDLOW = file.variables["CLDLOW"][:]
    # SAM's CLDLOW variable uses a threshold for cloud fraction that is based on LWP (20 g/m2, when CEM=.true., as in these runs)
    # which may not be consistent with an optical-depth-based threshold, as would be used in a satellite product.
    CLDMID = file.variables["CLDMID"][:]
    CLDHI = file.variables["CLDHI"][:]
    MODISREL = file.variables["MODISREL"][:]  # MODIS Effective Radius (Liquid): mkm
    ISCCPLOW = file.variables["ISCCPLOW"][:]
    # The ISCCP simulator approximates the cloud fraction that would be observed by a satellite using an optical depth threshold
    # of 0.3. The model domain includes only the lower troposphere, so that the relevant ISCCPTOT and ISCCPLOW should be identical.
    ISCCPTOT = file.variables["ISCCPTOT"][:]  #
    ZINV_org = file.variables["ZINV"][:]  # inversion height (km)
    WMAX = file.variables["WMAX"][:]  # max updraft vel. (m/s)
    LWNTOA = file.variables["LWNTOA"][:]  # Net LW flux at TOA (w/m2)
    SWNTOA = file.variables["SWNTOA"][:]  # Net SW flux at TOA (w/m2)
    SOLIN = file.variables["SOLIN"][:]  # Incoming SW flux at TOA (w/m2)
    ISCCPALB = file.variables["ISCCPALB"][:]  # ISCCP Cloud Albedo
    SWNTOAC = file.variables["SWNTOAC"][:]  # Net SW flux at TOA (Clear Sky)
    LWNTOAC = file.variables["LWNTOAC"][:]  # Net LW flux at TOA (Clear Sky)
    MODISTAU = file.variables["MODISTAU"][:]  # MODIS Cloud Optical Path
    ISCCPTAU = file.variables["ISCCPTAU"][:]  # ISCCP Optical Path
    TAUQCacc = file.variables["TAUQC"][:]  # Approx optical depth of cloud liquid water
    TAUQRacc = file.variables["TAUQR"][:]  # Approx optical depth of RAIN
    QCOEFFR = file.variables["QCOEFFR"][
        :
    ]  # Mixing ratio of QC over effective radius, EFFR = QC/QCOEFFR: g/kg/micro

    time = TIME - int(TIME[0])
    SOLIN2 = SOLIN.copy()
    SOLIN2[SOLIN2 <= 100] = np.nan
    albedo = (SOLIN2 - SWNTOA) / SOLIN2
    albd_clear = (SOLIN2 - SWNTOAC) / SOLIN2
    SW_CRE = SWNTOA - SWNTOAC

    # 2-D variables:
    NAd = file.variables["NAd"][:]  # dry aerosol number concentration (#/mg)
    NC = file.variables["NCCLD"][:]  # cloud number concentration in cloud
    NCORIG = file.variables["NC"][:]  # cloud number concentration
    NR = file.variables["NRCLD"][:]  # RAIN NUMBER CONCENTRATION in cloud
    NRORIG = file.variables["NR"][:]  # RAIN NUMBER CONCENTRATION
    RHO = file.variables["RHO"][:]  # kg/m3
    CLD = file.variables["CLD"][:]  # cloud Fraction
    #    CLDCUMDN=file.variables['CLDCUMD1'][:] # Cumulative Shaded Cloud Fraction, Computed Downwards
    RWC = file.variables["QPCLD"][:]  # rain water and snow content: g/kg
    CWC = file.variables["QNCLD"][:]  # cloud water and cloud ice content in cloud: g/kg
    QCCLD = file.variables["QCCLD"][:]  # Cloud liquid water mixing ratio in cloud
    QCOND = file.variables["QCOND"][:]  # Total Condensate: g/kg
    WOBS = file.variables["WOBS"][:]  # large scale W: m/s
    TABS = file.variables["TABS"][:]  # Absolute temperature: K
    QT = file.variables["QT"][:]  # Total water (no rain/snow included): g/kg
    QV = file.variables["QV"][:]  # Water vapor: g/kg
    RH = file.variables["RELH"][:]  # RH: %
    THETA = file.variables["THETA"][:]  # THETA: K
    U = file.variables["U"][:]  # x wind component: m/s
    V = file.variables["V"][:]  # y wind component: m/s
    P = file.variables["p"][:]  # pressure
    PRECIP = file.variables["PRECIP"][:]  # precipitation flux: mm/day
    W2 = file.variables["W2"][:]  # Variance of the z wind component: m2/s2
    THETAL = file.variables["THETAL"][:]  # Liquid water potential temperature: K

    QAd = file.variables["QAd"][:]  # DRY AEROSOL MASS: g/kg
    QAw = file.variables["QAw"][:]  # WET AEROSOL MASS
    QCORIG = file.variables["QC"][:]  # Cloud liquid water mass mixing ratio: g/kg
    QRORIG = file.variables["QR"][:]  # RAIN
    QAr = file.variables["QAr"][:]  # RAIN AEROSOL MASS

    # Aerosol budget terms:
    # NAd:
    # NC:
    NCSTRG = file.variables["NCSTRG"][
        :
    ]  # Storage of Cloud Water NUMBER CONCENTRATION: #/mg/day
    NCADV = file.variables["NCADV"][
        :
    ]  # Tendency of Cloud Water NUMBER CONCENTRATION due to resolved vertical advection: #/mg/day
    NCDIFF = file.variables["NCDIFF"][
        :
    ]  # Tendency of Cloud Water NUMBER CONCENTRATION due to vertical SGS transport: #/mg/day
    NCLSADV = file.variables["NCLSADV"][
        :
    ]  # Tendency of Cloud Water NUMBER CONCENTRATION due to large-scale vertical advecti: #/mg/day
    NCSED = file.variables["NCSED"][
        :
    ]  # Tendency of Cloud Water NUMBER CONCENTRATION due to sedimentation: #/mg/day
    NCMPHY = file.variables["NCMPHY"][
        :
    ]  # Tendency of CLOUD WATER NUMBER CONCENTRATION due to microphysical processes: #/mg/day
    NCFLXS = file.variables["NCFLXS"][
        :
    ]  # Subgrid flux of CLOUD WATER NUMBER CONCENTRATION: #/mg/day
    # NAd:
    try:
        NAdSTRG = (
            file.variables["NAcSTRG"][:] - NCSTRG
        )  # Storage of DRY AEROSOL NUMBER CONCENTRATION: #/mg/day
        NAdADV = (
            file.variables["NAcADV"][:] - NCADV
        )  # Tendency of DRY AEROSOL NUMBER CONCENTRATION due to resolved vertical advection: #/mg/day
        NAdDIFF = (
            file.variables["NAcDIFF"][:] - NCDIFF
        )  # Tendency of DRY AEROSOL NUMBER CONCENTRATION due to vertical SGS transport: #/mg/day
        NAdLSADV = (
            file.variables["NAcLSADV"][:] - NCLSADV
        )  # Tendency of DRY AEROSOL NUMBER CONCENTRATION due to large-scale vertical advecti: #/mg/day
        NAdSED = (
            file.variables["NAcSED"][:] - NCSED
        )  # Tendency of DRY AEROSOL NUMBER CONCENTRATION due to sedimentation: #/mg/day
        NAdMPHY = (
            file.variables["NAcMPHY"][:] - NCMPHY
        )  # Tendency of DRY AEROSOL NUMBER CONCENTRATION due to microphysical processes: #/mg/day
        NAdFLXS = (
            file.variables["NAcFLXS"][:] - NCFLXS
        )  # Subgrid flux of DRY AEROSOL NUMBER CONCENTRATION: #/m2/s
    except BaseException:
        NAdSTRG = file.variables["NAdSTRG"][
            :
        ]  # Storage of DRY AEROSOL NUMBER CONCENTRATION: #/mg/day
        NAdADV = file.variables["NAdADV"][
            :
        ]  # Tendency of DRY AEROSOL NUMBER CONCENTRATION due to resolved vertical advection: #/mg/day
        NAdDIFF = file.variables["NAdDIFF"][
            :
        ]  # Tendency of DRY AEROSOL NUMBER CONCENTRATION due to vertical SGS transport: #/mg/day
        NAdLSADV = file.variables["NAdLSADV"][
            :
        ]  # Tendency of DRY AEROSOL NUMBER CONCENTRATION due to large-scale vertical advecti: #/mg/day
        NAdSED = file.variables["NAdSED"][
            :
        ]  # Tendency of DRY AEROSOL NUMBER CONCENTRATION due to sedimentation: #/mg/day
        NAdMPHY = file.variables["NAdMPHY"][
            :
        ]  # Tendency of DRY AEROSOL NUMBER CONCENTRATION due to microphysical processes: #/mg/day
        NAdFLXS = file.variables["NAdFLXS"][
            :
        ]  # Subgrid flux of DRY AEROSOL NUMBER CONCENTRATION: #/m2/s
    SCVTNADC = file.variables["SCVTNADC"][
        :
    ]  # Tendency of NAD due to interstitial scavenging by cloud: #/mg/day
    SCVTNADR = file.variables["SCVTNADR"][
        :
    ]  # Tendency of NAD due to interstitial scavenging by rain: #/mg/day
    # NR:
    NRSTRG = file.variables["NRSTRG"][
        :
    ]  # Storage of Rain NUMBER CONCENTRATION: #/mg/day
    NRADV = file.variables["NRADV"][
        :
    ]  # Tendency of Rain NUMBER CONCENTRATION due to resolved vertical advection: #/mg/day
    NRDIFF = file.variables["NRDIFF"][
        :
    ]  # Tendency of Rain NUMBER CONCENTRATION due to vertical SGS transport: #/mg/day
    NRLSADV = file.variables["NRLSADV"][
        :
    ]  # Tendency of Rain NUMBER CONCENTRATION due to large-scale vertical advecti: #/mg/day
    NRSED = file.variables["NRSED"][
        :
    ]  # Tendency of Rain NUMBER CONCENTRATION due to sedimentation: #/mg/day
    NRMPHY = file.variables["NRMPHY"][
        :
    ]  # Tendency of Rain NUMBER CONCENTRATION due to microphysical processes: #/mg/day
    NCSTEN = file.variables["NCSTEN"][
        :
    ]  # CHANGE IN CLOUD DROPLET NUMBER DUE TO SEDIMENTATION: #/mg/day
    NRSTEN = file.variables["NRSTEN"][
        :
    ]  # CHANGE IN RAIN NUMBER DUE TO SEDIMENTATION: #/mg/day
    NPRA_pos = file.variables["NPRA"][
        :
    ]  # CHANGE IN RAIN NUMBER DUE TO ACCRETION OF CLOUD DROPLETS: #/mg/day
    NPRC = file.variables["NPRC"][:]  # CHANGE NC AUTOCONVERSION DROPLETS: #/mg/day
    NPRC1 = file.variables["NPRC1"][:]  # CHANGE NR AUTOCONVERSION DROPLETS: #/mg/day
    NCPOSLIM = file.variables["NCPOSLIM"][
        :
    ]  # CHANGE IN CLOUD DROPLET NUMBER DUE TO POSITIVE LIMITING: #/mg/day
    NCNEGLIM = file.variables["NCNEGLIM"][
        :
    ]  # CHANGE IN CLOUD DROPLET NUMBER DUE TO NEGATIVE LIMITING: #/mg/day
    NRPOSLIM = file.variables["NRPOSLIM"][
        :
    ]  # CHANGE IN RAIN DROPLET NUMBER DUE TO POSITIVE LIMITING: #/mg/day
    NRNEGLIM = file.variables["NRNEGLIM"][
        :
    ]  # CHANGE IN RAIN DROPLET NUMBER DUE TO NEGATIVE LIMITING: #/mg/day

    time_mg, z_mg = np.meshgrid(time, z)  # making meshgrid

    cld = CLDLOW.copy()
    cld[:] = np.nan
    for i in range(len(CLDLOW)):
        cld[i] = np.max(CLDLOW[i] + CLDMID[i] + CLDHI[i])

    # calculate accumulated precipitation
    acc_prec = PREC.copy()
    acc_prec[:] = np.nan
    for i in range(len(acc_prec)):
        acc_prec[i] = np.nansum(PREC[: i + 1]) / (time[-1] * 24)

    # calculate surface wind magnitude
    sfc_wnd = (U[:, 0] ** 2 + V[:, 0] ** 2) ** 0.5

    # SAM outputs only give the heights of cell centers.
    # To compute a vertical integral, you need the cell interface heights (i.e., the w levels):
    nz = len(z)
    zi = np.arange(nz + 1).astype(float)
    zi[:] = np.nan
    zi[0] = 0
    zi[1:nz] = 0.5 * (z[: nz - 1] + z[1:nz])
    zi[nz] = 1.5 * z[nz - 1] - 0.5 * z[nz - 2]
    dz = zi[1:] - zi[:-1]  # Compute the cell thicknesses

    ## Calc. inversion height
    # ZINV = inv_h(THETAL, RH, dz, z)
    dz3 = z.copy()
    dz3[:] = np.nan
    dz3[1:] = z[1:] - z[:-1]
    ZINV = inv_h(THETAL, RH, dz3, z)
    d_THETA = THETA[:, 1:] - THETA[:, :-1]

    #################################
    ### calculate modeled CTH
    # correct method
    #    SAM_CTH = CWP.copy()
    #    SAM_CTH[:] = np.nan
    #    for i in range(len(CWP)):
    #        if np.nanmax(CLDCUMDN[i,:]) > 0:
    #            tmp = CLDCUMDN[i,:] / np.nanmax(CLDCUMDN[i,:]) # normalize cumulative cloud fraction by its maxmimum value
    #            idx = np.nanmax(np.where(tmp > 0.5)[0]) # median CTH: height where normalized cumulative cloud fraction crosses 0.5
    #            SAM_CTH[i] = z[idx]

    # Calculate the index of BL:
    # method1: find index where z is equal to ZINV
    inv_idx = np.arange(len(ZINV))
    inv_idx[:] = int(0)
    for i in range(len(ZINV)):
        diff = abs(z - ZINV[i])
        inv_idx[i] = np.where(diff == np.min(diff))[0][0]
    # method2: find infrc where RH goes below 75%
    inv_idx2 = np.arange(len(ZINV))
    inv_idx2[:] = int(0)
    for i in range(len(ZINV)):
        j = 0
        while RH[i, j] > 75:
            j += 1
        inv_idx2[i] = j
    # Not very helpful method!

    # Calc. entrainment:
    entrainment = []  # np.arange(len(ZINV)) #entrainment[:] = int(0)
    for i in range(len(ZINV)):
        entrainment.append(-WOBS[i, inv_idx[i]])
    entrainment = np.array(entrainment) * 1000  # convert to mm/s

    # Marshall-Palmer and other R-Z relations
    #    Z = 10 ** (rad_refl_all[:,0] / 10)
    #    R_MP = (Z / 200) ** (1/1.6) * 24
    #    acc_R_MP = calc_acc_R(R_MP)

    #    R_C2004 = (Z / 57) ** (1/1.1) * 24
    #    acc_R_C2004 = calc_acc_R(R_C2004)

    #### MBL integral of number concentrations
    NA = NAd + NCORIG + NRORIG
    # NA  = NAd + NC + NR

    integ_NA_BL = mbl_ave_N(NA, RHO, dz, inv_idx)
    integ_NC_BL = mbl_ave_N(NC, RHO, dz, inv_idx)
    # integ_NR_BL  = mbl_ave_N(NR,  RHO, dz, inv_idx)
    # You have z(1:nz) which are the heights of the cell centers.
    # Compute the interface heights between cells: zi(1) = 0; zi(2:nz) = 0.5*(z(1:nz-1)+z(2:nz)); zi(nz+1) = 1.5*z(nz) - 0.5*z(nz-1)
    # Compute the cell thicknesses: dz(1:nz) = zi(2:nz+1) - zi(1:nz)

    ####

    ### Entrainment and W_lg
    w_ls_inv = -entrainment
    winw = 2
    entr = ZINV.copy()
    entr[:] = np.nan
    for i in range(len(time)):
        t_entr = 1 / 3
        tind = np.where(abs(time - time[i]) < t_entr / 2)[0]
        # print(tind)
        p = np.polyfit(time[tind], ZINV[tind], 1)
        entr[i] = 1e3 * p[0] / 86400 - w_ls_inv[i]
    entr = smooth(entr, winw)
    mean_entr = "{:.2f}".format(np.nanmean(entr))

    #### Calculate turbulence w'2
    W2_Zinv_200 = W2[:, 0].copy()
    W2_Zinv_200[:] = np.nan
    for i in range(W2[:, 0].size):
        ix_w2 = np.where(abs(ZINV[i] - 200 - z) == np.nanmin(abs(ZINV[i] - 200 - z)))[0]
        W2_Zinv_200[i] = W2[i, ix_w2]

    ### Calculate COD and Re
    def calc_OD(var):
        OD = var[:, 0].copy()
        OD[:] = np.nan
        for i in range(len(OD)):
            idx = np.where(QCCLD[i, :] > 0.01)[0]
            try:
                OD[i] = TAUQCacc[i, np.nanmax(idx)] - TAUQCacc[i, np.nanmin(idx)]
            except BaseException:
                pass
        return OD

    rho_idx = 65
    COD = calc_OD(TAUQCacc)
    SAM_re = (3.0 / 2.0) * (CWP / ISCCPTOT) / (COD * RHO[:, rho_idx] / ISCCPTOT)

    #### Calculate budget terms for aerosol number concentration
    NASTOR = NAdSTRG + NCSTRG + NRSTRG
    NAEDDY = NAdADV + NCADV + NRADV + NAdDIFF + NCDIFF + NRDIFF
    NALSADV = NAdLSADV + NCLSADV + NRLSADV
    NASED = NCSTEN + NRSTEN  # NAdSED + NCSED + NRSED
    NAMPHY = NAdMPHY + NCMPHY + NRMPHY - NASED
    NASCAV = -SCVTNADC - SCVTNADR

    NARC = -NPRC - NPRC1
    NPRA = -NPRA_pos
    # d NA = Srf Ad - (ScvCld Ad + ScvRn Ad) - Accr C +  SlfC R - Fallout R + (NMT Ad + NMT C + NMT R)
    NASTOR[NASTOR > 1000] = np.nan
    NASTOR[NASTOR < -1000] = np.nan
    NARESID = NASTOR - NAEDDY - NALSADV - NASED - NAMPHY - NASCAV
    NARESID[NARESID > 1000] = np.nan
    NARESID[NARESID < -1000] = np.nan

    ####
    mass_1500 = np.arange(NAd[:, 0].size).astype(float)
    mass_1500[:] = np.nan
    for i in range(NAd[:, 0].size):
        mass_1500[i] = np.nansum(RHO[i, :115] * dz[:115])

    NASURF_test = (
        86400 * NAdFLXS[:, 0] / (mass_1500 * 1e6)
    )  # convert from #/m2/s to #/mg/day

    #########
    height_NAEDDY = height_Ave(NAEDDY)
    height_NALSADV = height_Ave(NALSADV)
    height_NASED = height_Ave(NASED)
    height_NASCAV = height_Ave(NASCAV)
    height_NARC = height_Ave(NARC)
    height_NPRA = height_Ave(NPRA)

    MBL_mass = np.arange(NAd[:, 0].size).astype(float)
    MBL_mass[:] = np.nan
    for i in range(NAd[:, 0].size):
        MBL_mass[i] = np.nansum(RHO[i, : inv_idx[i] + 5] * dz[: inv_idx[i] + 5])

    height_NASURF = (
        86400 * NAdFLXS[:, 0] / (MBL_mass * 1e6)
    )  # convert from #/m2/s to #/mg/day
    height_ENTRNMT = height_NALSADV + height_NAEDDY - height_NASURF

    # globals().update(locals())
    return (
        time,
        z,
        dz,
        RHO,
        CWP,
        RWP,
        ISCCPTOT,
        ZINV,
        entr,
        LWNTOA,
        SWNS,
        P,
        TABS,
        inv_idx,
        W2_Zinv_200,
        acc_prec,
        THETA,
        QT,
        integ_NA_BL,
        integ_NC_BL,
        COD,
        SAM_re,
        height_NARC,
        height_ENTRNMT,
        height_NASURF,
        height_NPRA,
        height_NASED,
        height_NASCAV,
        sfc_wnd,
        albedo,
        NA,
        NC,
        PREC,
        RH,
        albd_clear,
        SOLIN2,
        SW_CRE,
        SWNTOAC,
        PRECIP,
        CLD,
        W2,
        WOBS,
        QV,
    )


def delta(THETA):
    nz = THETA.shape[1]
    THETAi = (
        np.arange(THETA.shape[0] * (nz + 1))
        .reshape(THETA.shape[0], nz + 1)
        .astype(float)
    )
    THETAi[:] = np.nan
    THETAi[:, 0] = 0
    THETAi[:, 1:nz] = 0.5 * (THETA[:, : nz - 1] + THETA[:, 1:nz])
    THETAi[:, nz] = 1.5 * THETA[:, nz - 1] - 0.5 * THETA[:, nz - 2]
    d_THETA = THETAi[:, 1:] - THETAi[:, :-1]
    return d_THETA


def inv_h(THETA, RH, dzz, zz):
    #    d_THETA = delta(THETA)
    #    d_RH = delta(RH)
    d_THETA = THETA[:, 1:] - THETA[:, :-1]
    d_RH = RH[:, 1:] - RH[:, :-1]

    ZINV = d_THETA[:, 0].copy()
    ZINV[:] = np.nan
    for i in range(d_THETA.shape[0]):
        z = zz[i, :] if len(zz.shape) > 1 else zz
        dz = dzz[i, :] if len(zz.shape) > 1 else dzz
        dTHETA_dz = d_THETA[i, :] / dz[1:]
        dRH_dz = d_RH[i, :] / dz[1:]
        dTHETA_dz[dTHETA_dz < 0] = 0
        dRH_dz[dRH_dz > 0] = 0
        func = dTHETA_dz * dRH_dz
        if len(zz.shape) > 1:
            indx = np.max(np.where(func == np.nanmin(func))[0])
        else:
            indx = np.min(np.where(func == np.nanmin(func))[0])
        # approximate func as a parabola around the minimum and find the height where that parabola is minimized.
        # This will allow the inversion height to vary continuously as the input profiles change.
        # inversion_test is defined at midpoints of grid
        zavg = 0.5 * (z[:-1] + z[1:])
        rnge = range(indx - 1, indx + 2)
        # we define the parabola, converting from m to km.
        pp = np.polyfit(1e-3 * zavg[rnge], func[rnge], 2)
        # take the derivative of the parabola in coeffient space.
        pp_prime = np.array([2 * pp[0], pp[1]])  # this is its derivative
        # find the zero-crossing of the derivative. This is the inversion height in meters
        z_inv = -1e3 * pp_prime[1] / pp_prime[0]
        ZINV[i] = z_inv
    return ZINV


# Calc. LTS
def calc_LTS(THETA, P):
    indx = np.where((P <= 701) & (P >= 699))[0]

    theta_700 = np.nanmean(THETA[:, indx], 1)
    theta_sfc = THETA[:, 0]
    LTS = theta_700 - theta_sfc
    return LTS


def mbl_ave_N(NA, RHO, dz, inv_idx):
    #    integ_NA_BL = np.arange(NAd[:,0].size)
    integ_NA_BL = NA[:, 0].copy()
    integ_NA_BL[:] = np.nan
    for i in range(NA[:, 0].size):
        dz2 = dz.copy()
        mmm = np.where(np.isnan(NA[i, :]) == 1)[0]
        if len(mmm) > 0:
            dz2[mmm] = np.nan
        mmm2 = np.where(NA[i, :] <= 1)[0]
        if len(mmm2) > 0 and len(mmm2) != len(NA):
            dz2[mmm2] = np.nan
        integ_NA_BL[i] = np.nansum(
            RHO[i, : inv_idx[i]] * dz2[: inv_idx[i]] * NA[i, : inv_idx[i]]
        ) / np.nansum(RHO[i, : inv_idx[i]] * dz2[: inv_idx[i]])
    return integ_NA_BL


def mbl_ave_flight(RFin_alt, RFin_N, RFin_RHO, iii, ZINV):
    if iii >= len(ZINV):
        jjj = np.where(abs(RFin_alt - ZINV[-1]) == np.nanmin(abs(RFin_alt - ZINV[-1])))[
            0
        ][0]
    else:
        jjj = np.where(
            abs(RFin_alt - ZINV[iii]) == np.nanmin(abs(RFin_alt - ZINV[iii]))
        )[0][0]
    RFin_alt_diff = RFin_alt[:-1] - RFin_alt[1:]
    RFin_alt_diff2 = RFin_alt_diff.copy()
    mmm = np.where(np.isnan(RFin_N[1:]) == 1)[0]
    RFin_alt_diff2[mmm] = np.nan
    mmm2 = np.where(RFin_N[1:] <= 0.5)[0]
    RFin_alt_diff2[mmm2] = np.nan
    RFin_N_mean = np.nansum(
        RFin_RHO[jjj:] * RFin_N[jjj:] * RFin_alt_diff2[jjj - 1 :]
    ) / np.nansum(RFin_alt_diff2[jjj - 1 :] * RFin_RHO[jjj:])
    return RFin_N_mean


def mbl_ave_MERRA(MERRA_Na, ZINV, MERRA_H, iii_MERRA, MERRA_RHO):
    MERRA_Na_mean2 = MERRA_Na[:, 0].copy()
    MERRA_Na_mean2[:] = np.nan
    for i in range(MERRA_Na_mean2.size):
        MERRA_Na_mean2[i] = mbl_ave_flight(
            MERRA_H[i, :], MERRA_Na[i, :], MERRA_RHO[i, :], iii_MERRA[i], ZINV
        )
    return MERRA_Na_mean2


def mbl_ave_Forcing(z_forc, MERRA_Na, Forcing_RHO, iii_Forcing, ZINV):
    MERRA_Na_mean2 = MERRA_Na[:, 0].copy()
    MERRA_Na_mean2[:] = np.nan
    for i in range(MERRA_Na_mean2.size):
        MERRA_Na_mean2[i] = mbl_ave_flight(
            z_forc[i, :, 0, 0], MERRA_Na[i, :], Forcing_RHO[i, :], iii_Forcing[i], ZINV
        )
    return MERRA_Na_mean2


def smooth(y, box_pts):
    box = np.ones(box_pts) / box_pts
    y_smooth = np.convolve(y, box, mode="same")
    return y_smooth



import matplotlib.pyplot as plt


def cre_timeseries(res, time, lb, upbound, init_time_all, lngh=5):
    xlim = time[-1]

    fig, axis = plt.subplots(2, 4, figsize=(25, 11))
    i = 0
    j = 0

    r_positive = res.loc[
        (slice(None), slice(None), slice(None), slice(None), slice(None), 1)
    ]
    r_negative = res.loc[
        (slice(None), slice(None), slice(None), slice(None), slice(None), -1)
    ]

    assert len(r_positive) == len(r_negative), "Datasets must have same length."

    for ii in range(len(r_positive)):
        rp = r_positive.iloc[ii]
        rn = r_negative.iloc[ii]
        j = ii % 4 if ii <= 6 else 3
        i = 0 if ii <= 3 else 1

        if j == 0 and i == 0:
            axis[i, j].set_ylabel("L06\n$ΔR$ ($W$ $m^{-2}$)", fontsize="20")
        elif j == 0 and i == 1:
            axis[i, j].set_ylabel("L10\n$ΔR$ ($W$ $m^{-2}$)", fontsize="20")

        (
            rN,
            rL,
            rC_nan_fd,
            CRE_M,
            CRE_T,
            CRE_L,
            CRE_CF,
            CRE_rs,
            A_M,
            A_T,
            A_L,
            A_CF,
            A_rs,
            CRE1,
            CRE2,
            CF1,
        ) = rp

        axis[i, j].fill_between(time, -upbound, upbound, color="k", alpha=0.075)
        axis[i, j].set_xlabel(
            "Time (days since " + init_time_all[ii] + ")", fontsize="20"
        )
        axis[i, j].set_xlim([0, xlim + xlim / lngh])
        axis[i, j].tick_params(axis="both", which="major", labelsize=16)
        axis[i, j].grid(linestyle=":", axis="y", linewidth=1)

        if i == 0:
            axis[i, j].set_ylim([-360, 70])
        else:
            axis[i, j].set_ylim([-205, 100])

        (
            rN,
            rL,
            rC_nan_fd,
            CRE_M2,
            CRE_T2,
            CRE_L2,
            CRE_CF2,
            CRE_rs2,
            A_M,
            A_T,
            A_L,
            A_CF,
            A_rs,
            CRE1_2,
            CRE2_2,
            CF2,
        ) = rn

        #####
        if i == 0 and j == 0:
            axis[i, j].set_xlim([0, xlim + xlim / lngh])
            axis[i, j].plot(
                time, (CRE_T + CRE_T2) / 2, linewidth=1, color="tab:red", label="CDNC"
            )
            axis[i, j].plot(
                time, (CRE_L + CRE_L2) / 2, linewidth=1, color="tab:green", label="LWP"
            )
            axis[i, j].plot(
                time, (CRE_CF + CRE_CF2) / 2, linewidth=1, color="k", label="CF"
            )
            axis[i, j].plot(
                time,
                (CRE_rs + CRE_rs2) / 2,
                linewidth=1,
                color="deepskyblue",
                label="residual",
            )
        else:
            axis[i, j].set_xlim([0, xlim + xlim / lngh])
            axis[i, j].plot(time, (CRE_T + CRE_T2) / 2, linewidth=1, color="tab:red")
            axis[i, j].plot(time, (CRE_L + CRE_L2) / 2, linewidth=1, color="tab:green")
            axis[i, j].plot(time, (CRE_CF + CRE_CF2) / 2, linewidth=1, color="k")
            axis[i, j].plot(
                time, (CRE_rs + CRE_rs2) / 2, linewidth=1, color="deepskyblue"
            )
        axis[i, j].plot(
            time,
            (CRE_M + CRE_M2) / 2,
            "-",
            linewidth=1,
            color="b",
            label="(" + lb[rp.name[1]] + ")" + " - " + "(" + lb[rp.name[0]] + ")",
        )
        axis[i, j].legend(fontsize="15", loc="lower right")

    plt.tight_layout(w_pad=1)
    return fig



import numpy as np  # linear algebra

import dcre.cre_decomposition as decomp  # noqa: E402
import dcre.erfani_2022.helpers as h  # noqa: E402


def quant_CRE_alltime_new(fn1, fn2, a_ft, box, factor=1, cre_scaling=False):
    (
        time,
        z,
        dz,
        RHO,
        CWP,
        RWP,
        ISCCPTOT,
        ZINV,
        entr,
        LWNTOA,
        SWNS,
        P,
        TABS,
        inv_idx,
        W2_Zinv_200,
        acc_prec,
        THETA,
        QT,
        integ_NA_BL,
        integ_NC_BL,
        COD,
        SAM_re,
        height_NARC,
        height_ENTRNMT,
        height_NASURF,
        height_NPRA,
        height_NASED,
        height_NASCAV,
        sfc_wnd,
        albedo,
        NA,
        NC,
        PREC,
        RH,
        albd_clear,
        SOLIN2,
        SW_CRE,
        SWNTOAC,
        PRECIP,
        CLD,
        W2,
        WOBS,
        QV,
    ) = h.read_model_vars(fn1)

    # CF1  = smooth(ISCCPTOT, 10)
    # CF1 = ISCCPTOT
    #     T_ft1= SWNTOAC / SOLIN2  # old method: transmissivity of the free troposphere
    RADSWDN = fn1.variables["RADSWDN"][:]  # Downward shortwave radiative flux: W/m2
    RADSWDN_Zinv = SOLIN2.copy()
    RADSWDN_Zinv[:] = np.nan
    for ij in range(len(RADSWDN_Zinv)):
        RADSWDN_Zinv[ij] = RADSWDN[ij, inv_idx[ij] + 5]
    T_ft1 = RADSWDN_Zinv / SOLIN2  # new method: transmissivity of the free troposphere

    N1 = np.concatenate([integ_NC_BL[:3], h.smooth(integ_NC_BL, box)[3:]])
    L1 = h.smooth(CWP + RWP, 10)
    simulation1 = decomp.Simulation(
        CWP=L1,
        CF=ISCCPTOT,
        inv_idx=inv_idx,
        NCCN=N1,
        albedo=albedo,
        albd_clear=albd_clear,
        SOLIN=SOLIN2,
        sw_dwn_profile=RADSWDN_Zinv,
    )

    #####
    (
        _,
        _,
        _,
        _,
        CWP,
        RWP,
        ISCCPTOT,
        _,
        _,
        _,
        _,
        _,
        _,
        inv_idx,
        _,
        _,
        _,
        _,
        _,
        integ_NC_BL,
        _,
        _,
        _,
        _,
        _,
        _,
        _,
        _,
        _,
        albedo,
        _,
        _,
        _,
        _,
        albd_clear,
        SOLIN2,
        SW_CRE,
        _,
        _,
        _,
        _,
        _,
        _,
    ) = h.read_model_vars(fn2)

    RADSWDN = fn2.variables["RADSWDN"][:]  # Downward shortwave radiative flux: W/m2
    RADSWDN_Zinv = SOLIN2.copy()
    RADSWDN_Zinv[:] = np.nan
    for ij in range(len(RADSWDN_Zinv)):
        RADSWDN_Zinv[ij] = RADSWDN[ij, inv_idx[ij] + 5]
    T_ft2 = RADSWDN_Zinv / SOLIN2  # new method: transmissivity of the free troposphere

    N = np.concatenate([integ_NC_BL[:3], h.smooth(integ_NC_BL, box)[3:]])
    L = h.smooth(CWP + RWP, 10)
    simulation2 = decomp.Simulation(
        CWP=L,
        CF=ISCCPTOT,
        inv_idx=inv_idx,
        NCCN=N,
        albedo=albedo,
        albd_clear=albd_clear,
        SOLIN=SOLIN2,
        sw_dwn_profile=RADSWDN_Zinv,
    )

    CRE_Decomposer = decomp.CRE_Decompositer(simulation1, simulation2)
    CRE_Decomposer.decompose(factor=factor, cre_scaling=cre_scaling)

    return CRE_Decomposer.return_erfani_2022()
